"""
.. include:: ../../README.md
"""

from .base import AnyFeature as AnyFeature
from .base import BaseFeature as BaseFeature
from .base import Info as Info
from .decorator import feature as feature
from .decorator import input_feature as input_feature
from .resolution import Resolution as Resolution
from .resolution import Trippa as Trippa
from .results import FeatureResult as FeatureResult
